// recfilter - examine Received: headers looking for Received from yadda,
// yadda by my-mail-host and picking off the [nn.nn.nn.nn] ip address
// of the from part of the header string.
//
// Designed to be run in a filter chain following bogofilter so the
// bogofilter spam header can be used to automagically add IP addrs
// to the spam source database.
//
// HINT: May want to add new tables to database. Can find the list
// of tables already in database by running:
//
// "SELECT name FROM sqlite_master WHERE type='table'" query
// to get names of tables already defined and create any new tables
// needed in old databases.

#include "mail-message.h"
#include <iostream>
#include <sstream>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <string>
#include <vector>
#include <unistd.h>
#include <sqlite3.h>
#include <regex>

enum ProgramActions {
   PA_UNSPECIFIED,
   PA_ADDSERVER,        // --serverpat
   PA_DELSERVER,        // --delserver
   PA_DELIP,            // --delip
   PA_CLEAN,            // --clean
   PA_FILTER,           // --filter
   PA_LIST,             // --list
   PA_ADDWHITE,         // --addwhite
   PA_DELWHITE,         // --delwhile
};

static const char *
what_name(ProgramActions act) {
   switch(act) {
   case PA_ADDSERVER:   return "--serverpat";
   case PA_DELSERVER:   return "--delserver";
   case PA_DELIP:       return "--delip";
   case PA_CLEAN:       return "--clean";
   case PA_FILTER:      return "--filter";
   case PA_LIST:        return "--list";
   case PA_ADDWHITE:    return "--addwhite";
   case PA_DELWHITE:    return "--delwhite";
   default:             return "unknown";
   }
}

static ProgramActions what_to_do = PA_UNSPECIFIED;

static void open_database();
static void close_database();
static void add_server_names();
static void del_server_names();
static void add_whitelist_names();
static void del_whitelist_names();
static void del_ipaddrs();
static void clean_database();
static void list_database();
static std::string wildcard_to_regexp(const std::string str);
static void get_server_names();
static void get_whitelist();
static bool check_spam_ip(std::string);
static void record_spam_ip(std::string);
static void database_help();
static void server_help();
static void delserver_help();
static void delip_help();
static void clean_help();
static void help();
static void say_what(ProgramActions);
static bool table_exists(const char * tabname);
static void addwhite_help();
static void delwhite_help();

static std::string database_name;

static sqlite3 * db = NULL;

std::vector<std::string> server_strings;

std::vector<std::string> whitelist_strings;

std::vector<std::string> ip_strings;

unsigned long clean_days = 0;

int
main(int argc, char ** argv) {
   for (int i = 1; i < argc; ++i) {
      if (strncmp(argv[i], "--database=", 11) == 0) {
         if (strcmp(argv[i]+11, "help") == 0) {
            database_help();
            return 0;
         } else {
            database_name.append(argv[i]+11);
         }
      } else if (strncmp(argv[i], "--serverpat=", 12) == 0) {
         if (strcmp(argv[i]+12, "help") == 0) {
            server_help();
            return 0;
         } else {
            server_strings.push_back(argv[i]+12);
            say_what(PA_ADDSERVER);
         }
      } else if (strncmp(argv[i], "--delserver=", 12) == 0) {
         if (strcmp(argv[i]+12, "help") == 0) {
            delserver_help();
            return 0;
         } else {
            server_strings.push_back(argv[i]+12);
            say_what(PA_DELSERVER);
         }
      } else if (strncmp(argv[i], "--delip=", 8) == 0) {
         if (strcmp(argv[i]+8, "help") == 0) {
            delip_help();
            return 0;
         } else {
            ip_strings.push_back(argv[i]+8);
            say_what(PA_DELIP);
         }
      } else if (strncmp(argv[i], "--clean=", 8) == 0) {
         if (strcmp(argv[i]+8, "help") == 0) {
            clean_help();
            return 0;
         } else {
            char * endp;
            clean_days = strtoul(argv[i]+8, &endp, 10);
            if (*endp != '\0') {
               std::cerr << argv[i]+8 << " is not a valid number." << std::endl;
               return 2;
            }
            say_what(PA_CLEAN);
         }
      } else if (strcmp(argv[i], "--filter") == 0) {
         say_what(PA_FILTER);
      } else if (strcmp(argv[i], "--list") == 0) {
         say_what(PA_LIST);
      } else if (strncmp(argv[i], "--addwhite=", 11) == 0) {
         if (strcmp(argv[i]+11, "help") == 0) {
            addwhite_help();
            return 0;
         } else {
            whitelist_strings.push_back(argv[i]+11);
            say_what(PA_ADDWHITE);
         }
      } else if (strncmp(argv[i], "--delwhite=", 11) == 0) {
         if (strcmp(argv[i]+11, "help") == 0) {
            delwhite_help();
            return 0;
         } else {
            whitelist_strings.push_back(argv[i]+11);
            say_what(PA_DELWHITE);
         }
      } else if (strncmp(argv[i], "--help", 6) == 0) {
         help();
         return 0;
      } else {
         std::cerr << "Unrecognized argument: " << argv[i] << std::endl;
         std::cerr << "Try using --help option" << std::endl;
         return 2;
      }
   }

   if (what_to_do == PA_UNSPECIFIED) {
      say_what(PA_FILTER);
   }

   if (database_name.empty()) {
      const char * home = getenv("HOME");
      if ((home == NULL) || (strlen(home) == 0)) {
         std::cerr << "$HOME environment variable is not set." << std::endl;
         return 2;
      }
      database_name.append(home);
      if (database_name[database_name.size()-1] != '/') {
         database_name.append("/");
      }
      database_name.append(".recfilter");
   }

   open_database();

   switch(what_to_do) {
   case PA_ADDSERVER:
      add_server_names();
      close_database();
      return 0;
   case PA_DELSERVER:
      del_server_names();
      close_database();
      return 0;
   case PA_ADDWHITE:
      add_whitelist_names();
      close_database();
      return 0;
   case PA_DELWHITE:
      del_whitelist_names();
      close_database();
      return 0;
   case PA_DELIP:
      del_ipaddrs();
      close_database();
      return 0;
   case PA_CLEAN:
      clean_database();
      close_database();
      return 0;
   case PA_LIST:
      list_database();
      close_database();
      return 0;
   default:
      break;
   }

   // get server name patterns from database.

   get_server_names();

   // get whitelisted IP addr patterns from database.

   get_whitelist();

   // Read mail message from cin and get header

   MailMessage mm;
   std::vector<std::string> & recs = mm.Received();
   int numrec = recs.size();
   std::string ipaddr;
   bool is_bogus = false;

   // Look through all the received headers...

   for (int r = numrec-1; (r >= 0) && (ipaddr.size() == 0); --r) {
      std::string t = recs[r];
      int numserv = server_strings.size();
      for (int s = 0; s < numserv; ++s) {
         std::regex spat(wildcard_to_regexp(server_strings[s]));
         std::smatch m;
         if (std::regex_search(t, m, spat) && (m.size() > 0)) {

            // Found one saying "by my.server.name", look for "from" IP
            // somewhere in string before "by" position.

            int bypos = m.position(0);

            int frompos = t.find("from");
            if ((frompos >= 0) && (frompos < bypos)) {
               int bpos = t.find_first_of("[(", frompos);
               if ((bpos > frompos) && (bpos < bypos)) {
                  int epos = t.find_first_of(")]", bpos);
                  if ((epos > bpos) && (epos < bypos)) {
                     bpos = t.find_first_not_of("[(", bpos);
                     if ((bpos > frompos) && (bpos < epos)) {
                        std::string testaddr = t.substr(bpos, epos-bpos);
                        int brack = testaddr.find("[");
                        if (brack >= 0) {
                           testaddr = testaddr.substr(brack+1);
                        }
                        if (testaddr != "127.0.0.1") {

                           ipaddr = testaddr;

                           // Found a bracketed IP addr that is not
                           // localhost, this must be where the mail really
                           // came from.  Has bogofilter already marked this
                           // as spam?

                           std::string bogo = mm.Bogosity();
                           if (bogo.find("Spam,") == 0) {
                              is_bogus = true;
                           }
                           break;
                        }
                     }
                  }
               }
            }
         }
      }
   }
   if (ipaddr.size() > 0) {
      int numwhite = whitelist_strings.size();
      for (int w = 0; w < numwhite; ++w) {
         std::regex wpat(wildcard_to_regexp(whitelist_strings[w]));
         std::smatch m;
         if (std::regex_match(ipaddr, m, wpat)) {
            // This ipaddr matches a whitelist entry. Forget we saw it.
            ipaddr.clear();
            break;
         }
      }
   }
   if (ipaddr.size() > 0) {
      if (! is_bogus) {
         is_bogus = check_spam_ip(ipaddr);
      }
      if (is_bogus) {
         record_spam_ip(ipaddr);
         mm.SetSpamIP(ipaddr);
      }
   }
   close_database();

   mm.WriteToStdout();

   return 0;
}

static const char initialize_sql [] =
   "create table ipaddrs ( ipaddr VARCHAR(15) PRIMARY KEY, recent DATE );"
   "create table serverpats ( server VARCHAR(100) PRIMARY KEY );"
;

static void open_database() {
   if (db != NULL) {
      // Already opened database.
      return;
   }

   // Remember if the database existed before we opened it.

   bool need_to_init = access(database_name.c_str(), R_OK|W_OK) != 0;

   if (sqlite3_open_v2(database_name.c_str(),
                       &db,
                       SQLITE_OPEN_READWRITE|SQLITE_OPEN_CREATE,
                       NULL)) {
      std::cerr << "Cannot open database \""
                << database_name
                << "\": "
                << sqlite3_errmsg(db)
                << std::endl;
      sqlite3_close(db);
      exit(2);
   }

   char * errmsg;

   // Don't get busy errors if multiple runs happen to occur at the same
   // time, just wait up to a couple of seconds (which should be enough for
   // any one run to finish).

   sqlite3_busy_timeout(db, 2000);

   // Start a transaction that covers everything trollfilter does during
   // this run.

   if (sqlite3_exec(db, "begin;", NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Error starting transaction: "
                << errmsg
                << std::endl;
      sqlite3_free(errmsg);
   }

   if (need_to_init) {
      if (sqlite3_exec(db, initialize_sql, NULL, 0, &errmsg) != SQLITE_OK) {
         std::cerr << "Unable to create required tables: "
                   << errmsg
                   << std::endl;
         sqlite3_free(errmsg);
         close_database();
         exit(2);
      }
   }
   if (! table_exists("whitelist")) {
      if (sqlite3_exec(db, 
             "create table whitelist ( ipaddr VARCHAR(100) PRIMARY KEY );",
             NULL, 0, &errmsg) != SQLITE_OK) {
         std::cerr << "Unable to create whitelist table: "
                   << errmsg
                   << std::endl;
         sqlite3_free(errmsg);
         close_database();
         exit(2);
      }
   }
}

static void close_database() {
   if (db == NULL) {
      return;
   }

   char * errmsg;

   // Commit the transaction.

   if (sqlite3_exec(db, "commit;", NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Error committing transaction: "
                << errmsg
                << std::endl;
      sqlite3_free(errmsg);
   }

   sqlite3_close(db);
   db = NULL;
}

static void add_server_names() {
   int len = server_strings.size();
   for (int i = 0; i < len; ++i) {
      std::string cmd("insert or ignore into serverpats (server) values ('");
      cmd.append(server_strings[i]);
      cmd.append("');");
      char * errmsg;
      if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
         std::cerr << "Adding serverpat error: " << errmsg << std::endl;
         sqlite3_free(errmsg);
      }
   }
}

static void del_server_names() {
   int len = server_strings.size();
   for (int i = 0; i < len; ++i) {
      std::string cmd("delete from serverpats where server='");
      cmd.append(server_strings[i]);
      cmd.append("';");
      char * errmsg;
      if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
         std::cerr << "Deleting serverpat error: " << errmsg << std::endl;
         sqlite3_free(errmsg);
      }
   }
}

static void add_whitelist_names() {
   int len = whitelist_strings.size();
   for (int i = 0; i < len; ++i) {
      std::string cmd("insert or ignore into whitelist (ipaddr) values ('");
      cmd.append(whitelist_strings[i]);
      cmd.append("');");
      char * errmsg;
      if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
         std::cerr << "Adding whitelistpat error: " << errmsg << std::endl;
         sqlite3_free(errmsg);
      }
   }
}

static void del_whitelist_names() {
   int len = whitelist_strings.size();
   for (int i = 0; i < len; ++i) {
      std::string cmd("delete from whitelist where ipaddr='");
      cmd.append(whitelist_strings[i]);
      cmd.append("';");
      char * errmsg;
      if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
         std::cerr << "Deleting whitelistpat error: " << errmsg << std::endl;
         sqlite3_free(errmsg);
      }
   }
}

static void del_ipaddrs() {
   int len = ip_strings.size();
   for (int i = 0; i < len; ++i) {
      std::string cmd("delete from ipaddrs where ipaddr='");
      cmd.append(ip_strings[i]);
      cmd.append("';");
      char * errmsg;
      if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
         std::cerr << "Deleting ipaddr error: " << errmsg << std::endl;
         sqlite3_free(errmsg);
      }
   }
}

static std::string wildcard_to_regexp(const std::string str) {
   // A server name can have just a * in place of one component
   // of a dotted name. That * will match any component. We
   // implement that by converting the wildcard string to a
   // regular expression.

   std::string rval("\\b");
   std::string::const_iterator it;
   for (it = str.begin(); it != str.end(); ++it) {
      char c = *it;
      if (isalnum(c) || isspace(c) || (c == '_') || (c == '-')) {
         rval.append(1, c);
      } else if (c == '*') {
         rval.append(".*");
      } else {
         rval.append("\\");
         rval.append(1,c);
      }
   }
   rval.append("\\b");
   return rval;
}

static void get_server_names() {
   std::string cmd("select * from serverpats;");
   sqlite3_stmt * sp = NULL;
   if (sqlite3_prepare_v2(db, cmd.c_str(), cmd.size()+1, &sp, NULL)
       == SQLITE_OK) {
      while (sqlite3_step(sp) == SQLITE_ROW) {
         const unsigned char * addr = sqlite3_column_text(sp, 0);
         int addrsize = sqlite3_column_bytes(sp, 0);
         std::string saddr((const char *)addr, addrsize);
         server_strings.push_back(saddr);
      }
   }
   sqlite3_finalize(sp);
   sp = NULL;
}

static void get_whitelist() {
   std::string cmd("select * from whitelist;");
   sqlite3_stmt * sp = NULL;
   if (sqlite3_prepare_v2(db, cmd.c_str(), cmd.size()+1, &sp, NULL)
       == SQLITE_OK) {
      while (sqlite3_step(sp) == SQLITE_ROW) {
         const unsigned char * addr = sqlite3_column_text(sp, 0);
         int addrsize = sqlite3_column_bytes(sp, 0);
         std::string saddr((const char *)addr, addrsize);
         whitelist_strings.push_back(saddr);
      }
   }
   sqlite3_finalize(sp);
   sp = NULL;
}

static std::string ip3part(std::string ipaddr) {
   std::string rval;
   unsigned int dotpos = ipaddr.find_last_of('.');
   if (dotpos != std::string::npos) {
      rval = ipaddr.substr(0,dotpos);
   }
   return rval;
}

static bool check_one_ip(std::string ipaddr) {
   bool rval = false;
   std::string cmd("select * from ipaddrs where ipaddr='");
   cmd.append(ipaddr);
   cmd.append("';");
   sqlite3_stmt * sp = NULL;
   if (sqlite3_prepare_v2(db, cmd.c_str(), cmd.size()+1, &sp, NULL)
       == SQLITE_OK) {
      if (sqlite3_step(sp) == SQLITE_ROW) {

         // Finding a matching row means that the ipaddr is in table

         rval = true;
      }
   }
   sqlite3_finalize(sp);
   return rval;
}

static void record_one_ip(std::string ipaddr) {
   std::string cmd("insert or ignore into ipaddrs (ipaddr,recent) values ('");
   cmd.append(ipaddr);
   cmd.append("',DATETIME('NOW'));");
   char * errmsg;
   if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Adding ipaddr error: " << errmsg << std::endl;
      sqlite3_free(errmsg);
   }
   std::string
      cmd2("update ipaddrs set recent=DATETIME('NOW') where ipaddr='");
   cmd2.append(ipaddr);
   cmd2.append("';");
   if (sqlite3_exec(db, cmd2.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Updating ipaddr error: " << errmsg << std::endl;
      sqlite3_free(errmsg);
   }
}

static bool check_spam_ip(std::string ipaddr) {
   // Is ipaddr in the table?
   bool rval = check_one_ip(ipaddr);
   if (! rval) {
      std::string ip3 = ip3part(ipaddr);
      if (! ip3.empty()) {
         // How about the 1st 3 components of the ipaddr?
         rval = check_one_ip(ip3);
      }
   }

   // NOTE: This would be the ideal place to add RBL checks if I am about
   // to return false.

   return rval;
}

static void record_spam_ip(std::string ipaddr) {
   record_one_ip(ipaddr);
   std::string ip3 = ip3part(ipaddr);
   if (! ip3.empty()) {
      record_one_ip(ip3);
   }
}

static void clean_database() {

   // Delete all the IPADDRs with no activity for clean_days days

   std::stringstream ss;
   ss << clean_days;

   std::string cmd("delete from ipaddrs where recent < datetime('now', '-");
   cmd.append(ss.str());
   cmd.append(" day');");

   char * errmsg;
   if (sqlite3_exec(db, cmd.c_str(), NULL, 0, &errmsg) != SQLITE_OK) {
      std::cerr << "Removing old ipaddr error: " << errmsg << std::endl;
      sqlite3_free(errmsg);
   }
}

static void list_database() {
   bool first_server = true;
   std::string cmd("select * from serverpats;");
   sqlite3_stmt * sp = NULL;
   if (sqlite3_prepare_v2(db, cmd.c_str(), cmd.size()+1, &sp, NULL)
       == SQLITE_OK) {
      while (sqlite3_step(sp) == SQLITE_ROW) {
         const unsigned char * addr = sqlite3_column_text(sp, 0);
         int addrsize = sqlite3_column_bytes(sp, 0);
         std::string saddr((const char *)addr, addrsize);
         if (first_server) {
            std::cout << "Mail server patterns in database:" << std::endl;
            first_server = false;
         }
         std::cout << "   " << saddr << std::endl;
      }
   }
   sqlite3_finalize(sp);
   bool first_white = true;
   std::string cmd3("select * from whitelist;");
   sp = NULL;
   if (sqlite3_prepare_v2(db, cmd3.c_str(), cmd3.size()+1, &sp, NULL)
       == SQLITE_OK) {
      while (sqlite3_step(sp) == SQLITE_ROW) {
         const unsigned char * addr = sqlite3_column_text(sp, 0);
         int addrsize = sqlite3_column_bytes(sp, 0);
         std::string saddr((const char *)addr, addrsize);
         if (first_white) {
            std::cout << "Whitelisted IP addrs in database:" << std::endl;
            first_white = false;
         }
         std::cout << "   " << saddr << std::endl;
      }
   }
   sqlite3_finalize(sp);
   sp = NULL;
   bool first_ip = true;
   std::string cmd2("select ipaddr, "
                    "julianday('now')-julianday(recent) from ipaddrs "
                    "order by recent desc;");
   if (sqlite3_prepare_v2(db, cmd2.c_str(), cmd2.size()+1, &sp, NULL)
       == SQLITE_OK) {
      std::string last_time;
      while (sqlite3_step(sp) == SQLITE_ROW) {
         const unsigned char * addr = sqlite3_column_text(sp, 0);
         int addrsize = sqlite3_column_bytes(sp, 0);
         std::string saddr((const char *)addr, addrsize);
         double daysold = sqlite3_column_double(sp, 1);
         int value;
         std::string units;
         if ((daysold + 0.5) >= 1.0) {
            value = (int)(daysold + 0.5);
            units.append("day");
         } else {
            daysold *= 24.0;
            if ((daysold + 0.5) >= 1.0) {
               value = (int)(daysold + 0.5);
               units.append("hour");
            } else {
               daysold *= 60.0;
               value = (int)(daysold + 0.5);
               units.append("minute");
            }
         }
         if (first_ip) {
            std::cout << "IP addrs database:" << std::endl;
            first_ip = false;
         }
         if (value != 1) {
            units.append("s");
         }
         std::stringstream ss;
         ss << value << " " << units << " old";
         std::string this_time = ss.str();
         if (last_time == this_time) {
            this_time.clear();
         } else {
            last_time = this_time;
            saddr.append(16 - saddr.size(), ' ');
         }

         std::cout << "   " << saddr << this_time << std::endl;
      }
   }
   sqlite3_finalize(sp);
   sp = NULL;
}

static bool
table_exists(const char * tabname) {
   if (db == NULL) {
      return false;
   }
   bool rval = false;
   std::string cmd("SELECT name FROM sqlite_master WHERE"
                   " type='table' AND name='");
   cmd.append(tabname);
   cmd.append("';");
   sqlite3_stmt * sp = NULL;
   if (sqlite3_prepare_v2(db, cmd.c_str(), cmd.size()+1, &sp, NULL)
       == SQLITE_OK) {
      if (sqlite3_step(sp) == SQLITE_ROW) {

         // Finding a matching row means that the tabname is in table

         rval = true;
      }
   }
   sqlite3_finalize(sp);
   return rval;
}

static void
say_what(ProgramActions new_action) {
   if ((what_to_do == new_action) || (what_to_do == PA_UNSPECIFIED)) {
      what_to_do = new_action;
      return;
   }
   std::cerr << "Option " << what_name(new_action)
             << " cannot be used with "
             << what_name(what_to_do) << std::endl;
   exit(2);
}

static void
print_usage_lines(const char ** lines) {
   while (*lines != NULL) {
      std::cerr << *lines++ << std::endl;
   }
}

static const char * database_help_lines[] = {
"--database=file Name the database file to use (default name is",
"$HOME/.recfilter). The database will be created and initialized if it does",
"not exist.",
NULL
};

static void database_help() {
   print_usage_lines(database_help_lines);
}

static const char * server_help_lines[] = {
"--serverpat=string Describe the string to look for on Received: headers",
"which identifies the Received: header that identifies mail coming from",
"somewhere and entering the local mail server system. This is usually",
"something like \"by yadda.yadda.com\".",
NULL
};

static void server_help() {
   print_usage_lines(server_help_lines);
}

static const char * delserver_help_lines[] = {
"--delserver=string Delete a string previously added by --serverpat=string",
NULL
};

static void delserver_help() {
   print_usage_lines(delserver_help_lines);
}

static const char * delip_help_lines[] = {
"--delip=ipaddr Delete an IP address from the database.",
NULL
};

static void delip_help() {
   print_usage_lines(delip_help_lines);
}

static const char * addwhite_help_lines[] = {
"--addwhite=string Specify a wildcard pattern to match a local IP address.",
"This might be something like 192.168.0.* to identify mail originating on the",
"local network that should never be treated as external spam.",
NULL
};

static void addwhite_help() {
   print_usage_lines(addwhite_help_lines);
}

static const char * delwhite_help_lines[] = {
"--delwhite=string Delete a string previously added by --addwhite=string",
NULL
};

static void delwhite_help() {
   print_usage_lines(delwhite_help_lines);
}

static const char * clean_help_lines[] = {
"--clean=days Clean the database of all ipaddrs that have not had any",
"activity for the specified number of days.",
NULL
};

static void clean_help() {
   print_usage_lines(clean_help_lines);
}

static const char * help_lines[] = {
"Usage: recfilter [options...] < email > email",
"",
"recfilter normally reads an email message on standard in and writes a",
"modified message on standard out, possibly adding an X-External-Spam: header",
"containing the IP address of the external server the mail came from.",
"",
"If a message has already been marked as spam by bogofilter, then the",
"external IP is added to the database, as is the 1st 3 components",
"of the IP address.",
"",
"If the message has not been marked as spam, the IP address is checked",
"against the database to see if the same or similar IP has sent spam",
"recently.",
"",
"--database=file    Give the location of the database file.",
"--serverpat=str    Give the string to identify mail server in Received: header.",
"--delserver=str    Delete previously added serverpat string.",
"--delip=ipaddr     Delete ipaddr from database.",
"--clean=days       Clean database of ipaddrs this many days old.",
"--filter           Run as a filter (default action).",
"--list             List the ipaddrs and server patterns in the database.",
"--addwhite=str     Whitelist pattern for IP addrs which are OK.",
"--delwhite=str     Delete previously added whitelist pattern.",
"--help             Print this message and exit.",
"",
"You can give \"help\" as an argument to any of the options that take",
"arguments to get more information about that option.",
NULL
};

static void help() {
   print_usage_lines(help_lines);
}
